#!/usr/bin/env python
#
# PerfEval.py: hack to check RISC simulation performance results from logs
#
# author: Rainer Doemer
#
# modifications:
# 04/26/18 RD	added play_trace example
# 07/14/17 RD	expectations adjusted for RISC 0.4.0 on host 'delta'
# 04/27/17 RD	adjusted for examples/demo
# 04/25/17 RD	initial version (test/regression)

# imports
import sys
import os
import time

# classes:
class Perf:
    "performance values of an example"
    def __init__(self,h,u,s,e,c,x):
	self.h = h
	self.u = u
	self.s = s
	self.e = e
	self.c = c
	self.x = x

# checks to be performed
Checks = [
    ("play",
	("seq.c<=100","ooo.c<=300","syn.c<=200","npd.c<=300",
	              "ooo.c>=200","syn.c>=100","npd.c>=200",
	 "AboutEqual(ooo.c,npd.c)",
	 "seq.e<=25.","ooo.e<=11.","syn.e<=16.","npd.e<=11.")),
    ("play_fifo",
	("seq.c<=100","ooo.c<=200","syn.c<=200","npd.c<=200",
	              "ooo.c>=100","syn.c>=100","npd.c>=100",
	 "AboutEqual(syn.c,npd.c)", "AboutEqual(syn.c,ooo.c)",
	 "seq.e<=25.","ooo.e<=16.","syn.e<=16.","npd.e<=16.")),
    ("play_trace",
	("seq.c<=100","ooo.c<=200","syn.c<=200","npd.c<=200",
	              "ooo.c>=100","syn.c>=100","npd.c>=100",
	 "AboutEqual(syn.c,npd.c)", "AboutEqual(syn.c,ooo.c)",
	 "seq.e<=25.","ooo.e<=16.","syn.e<=16.","npd.e<=16.")),
    ("mandelbrot",
	("seq.c<=100","ooo.c<=800","syn.c<=800","npd.c<=800",
	              "ooo.c>=500","syn.c>=500","npd.c>=500",
	 "AboutEqual(syn.c,npd.c)", "AboutEqual(syn.c,ooo.c)",
	 "seq.e>=70.","ooo.e>=10.","syn.e>=10.","npd.e>=10.",
	 "seq.e<=80.","ooo.e<=20.","syn.e<=20.","npd.e<=20.")),
    ("mandelbrot_fifo",
	("seq.c<=100","ooo.c<=800","syn.c<=800","npd.c<=800",
	              "ooo.c>=500","syn.c>=500","npd.c>=500",
	 "AboutEqual(syn.c,npd.c)", "AboutEqual(syn.c,ooo.c)",
	 "seq.e>=70.","ooo.e>=10.","syn.e>=10.","npd.e>=10.",
	 "seq.e<=80.","ooo.e<=20.","syn.e<=20.","npd.e<=20."))]

# functions
def PrintLogLine(d,e):
    print("%s: %0.2f user %0.2f system %0.2f elapsed %d%% CPU %d exit: %s" %
	(d.h, d.u, d.s, d.e, d.c, d.x, e))

def AboutEqual(a,b,Pvar=0.10):
    return (a >= b*(1.0-Pvar)) and (a <= b*(1.0+Pvar))

def ClearlyGreater(a,b,Pvar=0.10):
    return (a > b*(1.0+Pvar))

#
# Action starts here!
#

# parse command line arguments

Cmd = os.path.basename(sys.argv[0])
if len(sys.argv) < 2:
    sys.stderr.write("%s: bad arguments!\n" % Cmd)
    sys.stderr.write("Usage: %s <log file>\n" % Cmd)
    sys.exit(10)

LogName = sys.argv[1]
sys.stderr.write("%s: Reading log file \"%s\"...\n" % (Cmd,LogName))

# read the log file

try:
    File = open(LogName, "r")
except IOError, Msg:
    sys.stderr.write("%s: Cannot open file %s for reading: %s" % (Cmd,LogName,Msg))
    sys.exit(10)
Lines = File.readlines();
File.close()
sys.stderr.write("%s: %d lines read.\n" % (Cmd,len(Lines)))

# parse the contents

Examples = [ ]
Dict = { }
for l in Lines:
    (Host,u,Usr,s,Sys,e,Ela,c,Cpu,x,Exi,CmdLine) = [t(s) for t,s in zip((str,float,str,float,str,float,str,str,str,int,str,str),l.split())]
#   print Host,u,s,e,c,x,CmdLine
    Host = Host[:-1]	# get rid of ':' in host name
    c = int(c[:-1])	# get rid of '%' in CPU utilization
    if CmdLine.startswith("./"):
	CmdLine = CmdLine[2:]
#   print Host,u,s,e,c,x,CmdLine
    if CmdLine in Examples:
	sys.stderr.write("%s: Multiple results for '%s'\n" % (Cmd,CmdLine))
    else:
	Examples.append(CmdLine)
	Dict[CmdLine] = (Host,u,s,e,c,x)
sys.stderr.write("%s: %d result records.\n" % (Cmd,len(Dict)))

# execute the checks

NumChecks = 0
NumTests = 0
NumWarnings = 0
for check in Checks:
    (Name,Tests) = check
    Example = Name+"_seq"
    if Example in Examples:
	Data = Dict[Example]
	seq = Perf(Data[0],Data[1],Data[2],Data[3],Data[4],Data[5])
    else:
	sys.stderr.write("%s: Skipping '%s', no data available!\n"
				% (Cmd,Example))
	continue
    Example = Name+"_ooo"
    if Example in Examples:
	Data = Dict[Example]
	ooo = Perf(Data[0],Data[1],Data[2],Data[3],Data[4],Data[5])
    else:
	sys.stderr.write("%s: Skipping '%s', no data available!\n"
				% (Cmd,Example))
	continue
    Example = Name+"_syn"
    if Example in Examples:
	Data = Dict[Example]
	syn = Perf(Data[0],Data[1],Data[2],Data[3],Data[4],Data[5])
    else:
	sys.stderr.write("%s: Skipping '%s', no data available!\n"
				% (Cmd,Example))
	continue
    Example = Name+"_npd"
    if Example in Examples:
	Data = Dict[Example]
	npd = Perf(Data[0],Data[1],Data[2],Data[3],Data[4],Data[5])
    else:
	sys.stderr.write("%s: Skipping '%s', no data available!\n"
				% (Cmd,Example))
	continue
#   print seq, ooo, syn, npd
    NumChecks += 1
    for Test in Tests:
	NumTests += 1
	if not eval(Test):
	    NumWarnings += 1
	    print "\nWARNING: for %s, test '%s' failed!" % (Name,Test)
	    PrintLogLine(seq, Name+"_seq")
	    PrintLogLine(ooo, Name+"_ooo")
	    PrintLogLine(syn, Name+"_syn")
	    PrintLogLine(npd, Name+"_npd")

# result

if NumWarnings:
    print "\nWARNING: %d tests failed!" % NumWarnings
sys.stderr.write("%s: %d examples checked, a total of %d tests performed.\n"
		% (Cmd, NumChecks, NumTests))
sys.stderr.write("%s: Done.\n" % Cmd)

sys.exit(NumWarnings)

# EOF
