#include <iostream>

#include "risc.h"
#include <list>

#include "earliest_execution_table.h"

void
get_all_instances_of(
    std::list<risc::InstanceTree> &instances,
    risc::InstanceTree root,
    risc::Definition::Type type,
    bool first);

int main(int argc, char* argv[])
{
  std::cout << std::endl
            << "start processing file: " __FILE__
            << " ..."
            << std::endl;

  risc::Design d(argc,argv);

  // Creating a segment graph
  risc::sg::SegmentGraph::FunctionCallBoundarySet
    function_call_boundary(risc::tools::get_sc_wait_funcs());

  risc::sg::SegmentGraph sg(function_call_boundary);



  //////////////////////////////////////////////////////////
  std::list<risc::InstanceTree> module_instances;
  get_all_instances_of(module_instances,
      d.get_root(),
      risc::Definition::ModuleInstanceType,
      true);

  //////////////////////////////////////////////////////////

  for(std::list<risc::InstanceTree>::iterator
      iter  = module_instances.begin();
      iter != module_instances.end();
      iter++) {

    risc::ModuleInstance *module_instance
      = dynamic_cast<risc::ModuleInstance*>((*iter).get_object());

    for(risc::ThreadVectorIter
        iter_thread  = module_instance->definition_->threads_.begin();
        iter_thread != module_instance->definition_->threads_.end();
        iter_thread++) {

      if((*iter_thread)->get_name() == "thread1" ) {

        AstDOTGeneration dotgen;
        dotgen.generate((*iter_thread)->get_ast_node(), (*iter_thread)->get_name());
      }

      sg.add_function(*iter_thread, *iter_thread, true);
    }
  }

  //sg.clean_graph();

  // Creating the event conflict table
  risc::sg::EventConflictTable
    event_conflict_table(sg, d.path_instance_mapper_, true);

  sg.print_graph("sg_hierarchy.dot");
  EarlistExecutionTable execution_table(sg, event_conflict_table);
  execution_table.print_table();

  for(int
      row = 0;
      row < event_conflict_table.size_of_conflict_table_;
      row++) {

    for(int
        column = 0;
        column < event_conflict_table.size_of_conflict_table_;
        column++) {

      if(event_conflict_table.conflict_table_[row][column]) {

        risc::sg::Time time_row
          = execution_table.get_min_time(
                execution_table.table_index_segment_lookup_[row]);

        risc::sg::Time time_column
          = execution_table.get_min_time(
                execution_table.table_index_segment_lookup_[column]);

        //if(time_row.one_delta_cycle_ahead(time_column)) {

          std::pair<int, int> seg_and_inst_id_column
            = event_conflict_table.index_to_segment_and_instance_id(column);

          std::pair<int, int> seg_and_inst_id_row
            = event_conflict_table.index_to_segment_and_instance_id(row);

          std::cout << "Adding "
                    << seg_and_inst_id_column.first
                    << " to "
                    << seg_and_inst_id_row.first
                    << std::endl;


          std::pair<risc::sg::EdgeDescriptor, bool> edge =
            boost::add_edge(
                event_conflict_table.
                  segment_id_to_vertex_descriptor(seg_and_inst_id_row.first),
                event_conflict_table.
                  segment_id_to_vertex_descriptor(seg_and_inst_id_column.first),
                sg.graph_);

          if(edge.second) {
            sg.graph_[edge.first].event_notification_edge_ = true;
          }
        //}
      }
    }
  }

  sg.print_graph("cg_hierarchy.dot");

  std::cout << "++++++++++++++++++++++++++++" << std::endl;
  execution_table.optimize_table();
  execution_table.print_table();

  std::cout << std::endl;
  std::cout << "Print table" << std::endl;

  std::cout << std::endl;
  event_conflict_table.print_conflict_table("event_conflicts.html");

  std::cout << "end processing file: " __FILE__
            << " ..."
            << std::endl
            << std::endl;

  return 0;
}

// Assumption: only channels in channels and modules in modules
void get_all_instances_of(std::list<risc::InstanceTree> &instances,
         risc::InstanceTree root, risc::Definition::Type type, bool first)
{
  risc::ModuleInstance *module_instance
    = dynamic_cast<risc::ModuleInstance*>(root.get_object());

  if(module_instance) {

    for(risc::ModuleInstanceVectorIter
        iter  = module_instance->definition_->sub_modules_.begin();
        iter != module_instance->definition_->sub_modules_.end();
        iter++) {

      risc::InstanceTree::Element sub_path;

      sub_path.first = module_instance->definition_;
      sub_path.second = *iter;

      root.push_back(sub_path);
      get_all_instances_of(instances, root, type, false);
      root.pop_back();
    }

    for(risc::HierarchicalChannelInstanceVectorIter
        iter  = module_instance->definition_->sub_hier_channel_.begin();
        iter != module_instance->definition_->sub_hier_channel_.end();
        iter++) {

      risc::InstanceTree::Element sub_path;

      sub_path.first = module_instance->definition_;
      sub_path.second = *iter;

      root.push_back(sub_path);
      get_all_instances_of(instances, root, type, false);
      root.pop_back();
    }

    if(type == risc::Definition::EventType) {
      for(risc::EventVectorIter
          iter  = module_instance->definition_->events_.begin();
          iter != module_instance->definition_->events_.end();
          iter++) {

        risc::InstanceTree temp = root;
        risc::InstanceTree::Element e(module_instance->definition_, *iter);
        temp.push_back(e);
        instances.push_back(temp);
      }
    }

    if(type == risc::Definition::ModuleInstanceType) {

      if(first) {
        instances.push_back(root);
      }

      for(risc::ModuleInstanceVectorIter
          iter  = module_instance->definition_->sub_modules_.begin();
          iter != module_instance->definition_->sub_modules_.end();
          iter++) {

        risc::InstanceTree temp = root;
        risc::InstanceTree::Element e(module_instance->definition_, *iter);
        temp.push_back(e);
        instances.push_back(temp);
      }
    }
  }
}

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
