#include "vertex_property_writer.h"

#include "boost/algorithm/string.hpp"
#include <string>

// RD: was probably meant for debugging, not needed for the user
//#define WRITE_COMMENT_AFTER_BOUNDARY_STATEMENT\

risc::sg::VertexPropertyWriter::VertexPropertyWriter(Graph& graph):
  graph_(graph)
{ }

void
risc::sg::VertexPropertyWriter::write_segment(
                            std::ostream &out, Segment &seg)
{
  
    out << "[shape=record,label=\"{ Segment ID: " << seg.id_;

      if(seg.expressions_.size() != 0) {
        SgNode *first_node = seg.expressions_.front().node_;
        if(first_node) {
          SgFunctionDefinition *func_def =
            SageInterface::getEnclosingFunctionDefinition(first_node);
          if(func_def) {
            SgClassDefinition *class_def =
              SageInterface::getEnclosingClassDefinition(func_def);
            if(class_def) {
              out << " [" << SageInterface::get_name(class_def)
                  << "::" << SageInterface::get_name(func_def) << "]";
            }
          }
        }
      }

    if(seg.boundary_node_) {
      // RD: one would expect this wait statement to be printed,
      // RD: but for some reason TS didn't; I'll break this tradition now
      std::string unparsed_stmt = seg.boundary_node_->unparseToString();
      replace_escape_characters(unparsed_stmt);

      std::string seg_pcp = write_pcp(seg.pcp_);
      std::string file_name =
        Rose::StringUtility::stripPathFromFileName(
            seg.boundary_node_->get_file_info()->get_raw_filename());
      out << "\\n" << file_name
          << ":" << seg.boundary_node_->get_file_info()->get_line()
          << seg_pcp
          << " " << unparsed_stmt;

#ifdef WRITE_COMMENT_AFTER_BOUNDARY_STATEMENT
      // Here we writing the comment which is associated with the boundary
      // in the segment head
      const SgLocatedNode *located_node =
        isSgLocatedNode(seg.boundary_node_);
      AttachedPreprocessingInfoType *comments =
        const_cast<SgLocatedNode*>(
            located_node)->getAttachedPreprocessingInfo();
      if(comments != NULL) {
        // We are iterating over all associated comments, however, we print
        // only the comment which is in the same line like the boundary
        // statement
        for(AttachedPreprocessingInfoType::iterator
            i  = comments->begin();
            i != comments->end();
            i++) {
          if(located_node->get_file_info()->get_line()
              == (*i)->getLineNumber()) {
            // parsing out the '//' at the beginning
            std::string str = (*i)->getString().substr(2,
                                                  (*i)->getString().length()-2);
            // deleting the line break break at the end
            str.erase(std::remove(str.begin(), str.end(), '\n'), str.end());
            out << "  " << str;
          }
        }
      }
#endif // WRITE_COMMENT_AFTER_BOUNDARY_STATEMENT
    }

      for(Segment::Expressions::iterator
          expr_iter  = seg.expressions_.begin();
          expr_iter != seg.expressions_.end();
          expr_iter++) {

        //std::string unparsed_stmt = (*expr_iter)->unparseToString();
        std::string unparsed_stmt = expr_iter->node_->unparseToString();
        replace_escape_characters(unparsed_stmt);

        std::string temp =
          Rose::StringUtility::stripPathFromFileName(
            (*expr_iter).node_->get_file_info()->get_raw_filename());

// Only for communication graphs
//        if(isSgFunctionCallExp((*expr_iter))) {
          out << "|" << temp << ":"
              //<< (*expr_iter)->get_file_info()->get_line() << " "
              << expr_iter->node_->get_file_info()->get_line();

          std::string pcp = write_pcp(expr_iter->pcp_);
          out << pcp << " " << unparsed_stmt << "\\l";
//        }
      }

    out << "}\"]";

    /*
      // This is an alternative output strategy. Basically, everything is
      // represented in a (HTML) table, however, in my point of view it is
      // incredible ugly!
      out << "  [" << std::endl;
      out << "  shape = none" << std::endl;
      out << "  label = <<table border=\"1\" cellspacing=\"0\">" << std::endl;
      out << "            <tr><td>" << "ID: " << v << "</td></tr>" << std::endl;

      for(Segment::Statements::iterator
          stmt_iter  = seg.statements_.begin();
          stmt_iter != seg.statements_.end();
          stmt_iter++) {

        std::string unparsed_stmt = (*stmt_iter)->unparseToString();

        replace_escape_characters(unparsed_stmt);

        out << "            <tr><td>" << unparsed_stmt << "</td></tr>"
                                                       << std::endl;
      }
      out << "          </table>>" << std::endl;
      out << "  ]" << std::endl;
    */
}

void
risc::sg::VertexPropertyWriter::write_segment_RW_mode(
                            std::ostream &out, Segment &seg)
{
    out << "[shape=record,label=\"{ Segment ID: " << seg.id_;

    if(seg.boundary_node_) {
      // RD: no need to print this here, wait events are already printed below
      //std::string unparsed_stmt = seg.boundary_node_->unparseToString();
      //replace_escape_characters(unparsed_stmt);
      //out << "\\n " << unparsed_stmt;

      std::string temp =
        Rose::StringUtility::stripPathFromFileName(
            seg.boundary_node_->get_file_info()->get_raw_filename());
      out << " (" << temp << ":"
          << seg.boundary_node_->get_file_info()->get_line() << ")";

#ifdef WRITE_COMMENT_AFTER_BOUNDARY_STATEMENT
      // Here we writing the comment which is associated with the boundary
      // in the segment head
      const SgLocatedNode *located_node =
        isSgLocatedNode(seg.boundary_node_);
      AttachedPreprocessingInfoType *comments =
        const_cast<SgLocatedNode*>(
            located_node)->getAttachedPreprocessingInfo();
      if(comments != NULL) {
        // We are iterating over all associated comments, however, we print
        // only the comment which is in the same line like the boundary
        // statement
        for(AttachedPreprocessingInfoType::iterator
            i  = comments->begin();
            i != comments->end();
            i++) {
          if(located_node->get_file_info()->get_line()
              == (*i)->getLineNumber()) {
            // parsing out the '//' at the beginning
            std::string str = (*i)->getString().substr(2,
                                                  (*i)->getString().length()-2);
            // deleting the line break break at the end
            str.erase(std::remove(str.begin(), str.end(), '\n'), str.end());
            out << "  " << str;
          }
        }
      }
#endif // WRITE_COMMENT_AFTER_BOUNDARY_STATEMENT
    } else { // not a boundary node

      if(seg.expressions_.size() != 0) {
        SgNode *first_node = seg.expressions_.front().node_;
        if(first_node) {
          SgFunctionDefinition *func_def =
            SageInterface::getEnclosingFunctionDefinition(first_node);
          if(func_def) {
            SgClassDefinition *class_def =
              SageInterface::getEnclosingClassDefinition(func_def);
            if(class_def) {
              out << " [" << SageInterface::get_name(class_def)
                  << "::" << SageInterface::get_name(func_def) << "]";
            }
          }
        }
      }
    }

    // RD: no need for this here
    // RD: (and for sure not only the first list item!)
    //if(!seg.pcp_.path_.empty()) {
    //  std::string temp(seg.pcp_.path_.front()->unparseToString());
    //  replace_escape_characters(temp);
    //  out << " -- " << temp;
    //}

        // printing events this segment is sensitive to
        for(std::list<SymbolWithPath>::iterator
            var_symbol_iter  = seg.wait_events_.begin();
            var_symbol_iter != seg.wait_events_.end();
            var_symbol_iter++) {
          out << "| (wait) "
              << var_symbol_iter->symbol_->get_declaration()->get_qualified_name().getString()
              << write_pcp(var_symbol_iter->pcp_)
              << "\\l";
        }

        // printing events this segment is notifying
        for(std::list<SymbolWithPath>::iterator
            var_symbol_iter  = seg.notify_events_.begin();
            var_symbol_iter != seg.notify_events_.end();
            var_symbol_iter++) {
          out << "| (notify) "
              << var_symbol_iter->symbol_->get_declaration()->get_qualified_name().getString()
              << write_pcp(var_symbol_iter->pcp_)
              << "\\l";
        }

        // printing write variables
        for(std::list<SymbolWithPath>::iterator
            var_symbol_iter  = seg.write_variables_.begin();
            var_symbol_iter != seg.write_variables_.end();
            var_symbol_iter++) {
          out << "| (W) "
              << var_symbol_iter->symbol_->get_declaration()->get_qualified_name().getString()
              << write_pcp(var_symbol_iter->pcp_)
              << "\\l";
        }

        // printing read variables
        for(std::list<SymbolWithPath>::iterator
            var_symbol_iter  = seg.read_variables_.begin();
            var_symbol_iter != seg.read_variables_.end();
            var_symbol_iter++) {
          out << "| (R) "
              << var_symbol_iter->symbol_->get_declaration()->get_qualified_name().getString()
              << write_pcp(var_symbol_iter->pcp_)
              << "\\l";
        }
    out << "}\"]";
}

void
risc::sg::VertexPropertyWriter::write_segment_PSG_mode(
  std::ostream &out, Segment &seg)
{
  if(seg.is_fake_segment){
    if(seg.non_defining_function_call_name_!="")
      out << "[shape=record,style=dashed,label=\"{ Node ID: " << seg.node_id_;
    else
      out << "[shape=record,style=dotted,label=\"{ Node ID: " << seg.node_id_;      
  }
  else out << "[shape=record,style=bold,label=\"{ Node ID: " << seg.node_id_;

  if(seg.boundary_node_) {

    std::string temp =
    Rose::StringUtility::stripPathFromFileName(
      seg.boundary_node_->get_file_info()->get_raw_filename());

    out << " (" << temp << ":"
    << seg.boundary_node_->get_file_info()->get_line() << ")";

      // Here we writing the comment which is associated with the boundary
      // in the segment head
    const SgLocatedNode *located_node =
    isSgLocatedNode(seg.boundary_node_);

    AttachedPreprocessingInfoType *comments =
    const_cast<SgLocatedNode*>(
      located_node)->getAttachedPreprocessingInfo();

    if(comments != NULL) {

        // We are iterating over all associated comments, however, we print
        // only the comment which is in the same line like the boundary
        // statement
      for(AttachedPreprocessingInfoType::iterator
        i  = comments->begin();
        i != comments->end();
        i++) {

        if(located_node->get_file_info()->get_line()
          == (*i)->getLineNumber()) {

            // parsing out the '//' at the beginning
          std::string str = (*i)->getString().substr(2,
            (*i)->getString().length()-2);

            // deleting the line break break at the end
          str.erase(std::remove(str.begin(), str.end(), '\n'), str.end());
          out << "  " << str;
        }
      }
    }
  } else { // not a boundary node

    SgNode *first_node = NULL;
    if(seg.expressions_.size() != 0) {

      first_node = seg.expressions_.front().node_;

      if(first_node) {
        SgFunctionDefinition *func_def =
        SageInterface::getEnclosingFunctionDefinition(first_node);

        if(func_def) {

          SgClassDefinition *class_def =
          SageInterface::getEnclosingClassDefinition(func_def);

          if(class_def) {

            out << " [" << SageInterface::get_name(class_def)
            << "::" << SageInterface::get_name(func_def) << "]";
          }
        }
      }
    }
  }

  // RD: no need for this here either
  // RD: (this appears to be never used for current cases)
  //if(!seg.pcp_.path_.empty()) {
  //  std::string temp(seg.pcp_.path_.front()->unparseToString());
  //  replace_escape_characters(temp);
  //  out << " -- " << temp;
  //}

  bool w = true;

  std::vector<std::string> write_var_list;
  std::vector<std::string> read_var_list;
  std::vector<std::string> rw_var_list;
  std::map<std::string,int> var_written;
  std::map<std::string,int> var_read;
  std::map<std::string,int> var_rw;

  for(std::list<SymbolWithPath>::iterator
    var_symbol_iter  = seg.write_variables_.begin();
    var_symbol_iter != seg.write_variables_.end();
    var_symbol_iter++) 
  {

    if(!tools::is_member_variable(var_symbol_iter->symbol_)){
      continue;
    } 
    std::string var_name = (((var_symbol_iter->symbol_)->get_declaration())->get_qualified_name()).getString ();
    if(var_written[var_name] == 0){
      write_var_list.push_back(var_name);
      var_written[var_name]=1;
    }
  }

  for(std::list<SymbolWithPath>::iterator
    var_symbol_iter  = seg.read_variables_.begin();
    var_symbol_iter != seg.read_variables_.end();
    var_symbol_iter++) 
  {

    if(!tools::is_member_variable(var_symbol_iter->symbol_)){
      continue;
    } 
    std::string var_name = (((var_symbol_iter->symbol_)->get_declaration())->get_qualified_name()).getString ();
    if(var_read[var_name] == 0){
      read_var_list.push_back(var_name);
      var_read[var_name]=1;
    }
  }

  for(std::vector<std::string>::iterator
    iter  = write_var_list.begin();
    iter != write_var_list.end();
    iter++) 
  {

    if(var_read[*iter]>0){
      rw_var_list.push_back(*iter);
      var_rw[*iter]=1;
    }
  }

  for(std::vector<std::string>::iterator
    iter  = rw_var_list.begin();
    iter != rw_var_list.end();
    iter++) 
  {        
    if(w) out << "\" + \n                     \"";
    out << "| (RW) "
    << *iter;
    out << "\\l";
  }

  for(std::vector<std::string>::iterator
    iter  = write_var_list.begin();
    iter != write_var_list.end();
    iter++) 
  {
    if(var_rw[*iter]==1) continue;

    if(w) out << "\" + \n                     \"";
    out << "| (W) "
    << *iter;
    out << "\\l";
  }

  for(std::vector<std::string>::iterator
    iter  = read_var_list.begin();
    iter != read_var_list.end();
    iter++) 
  {
    if(var_rw[*iter]==1) continue;

    if(w) out << "\" + \n                     \"";
    out << "| (R) "
    << *iter;
    out << "\\l";
  }


  std::map <SgVariableSymbol*,int> event_notified;
  std::map <SgVariableSymbol*,int> event_waited;

      //print wait and notify events
  for(std::list<SymbolWithPath>::iterator
    iter = seg.wait_events_.begin();
    iter != seg.wait_events_.end();
    iter++)
  {
    if( event_waited[(*iter).symbol_] > 0) continue;
    
    event_waited[(*iter).symbol_] = 1;

    if(w) out << "\" + \n                     \"";
    out << "| (wait_event) " << (((*iter).symbol_)->get_declaration()->
                                get_qualified_name()).getString ();
    out << "\\l";
    w=true;
  }

  for(std::list<SymbolWithPath>::iterator
    iter = seg.notify_events_.begin();
    iter != seg.notify_events_.end();
    iter++)
  {
    if( event_notified[(*iter).symbol_] > 0) continue;
    event_notified[(*iter).symbol_]=1;

    if(w) out << "\" + \n                     \"";

    out << "| (notify_event) " << (((*iter).symbol_)->get_declaration()->
                                  get_qualified_name()).getString ();
    out << "\\l";
    w=true;
  }

  //print meta data
  if(seg.is_fake_segment){
    if(seg.non_defining_function_call_name_!=""){
      out << "\" + \n                     \"";
      out << "| "
      << seg.non_defining_function_call_name_;
      out << "\\l";
    }
  }
  out << "}\"]";
  
  out << "\n                     /*\n";
  out << "                     Segment node id = " << seg.id_ << "\n";
  if(seg.function_name_entrance_!="")
    out << "                     starting segment for function: " << seg.function_name_entrance_ << "\n";
  if(seg.function_name_exit_!="")
    out << "                     leaving segment for function: " << seg.function_name_exit_ << "\n";

  if(seg.is_fake_segment){
    if(seg.non_defining_function_call_name_!=""){
      out << "                     Partial Segment Type A, contains function: " << seg.non_defining_function_call_name_ << "\n";
      out << "                     Port Call Function for Port: " << seg.pcp_port_name_ << "\n";
      out << "                     Unique Seg Array Name: " << seg.unique_port_call_name_ << "\n";
    }
    else
      out << "                     Partial Segment Type B, will be merged into other segments during integration\n";
    
  }

  if(seg.is_simulation_process_)
    out << "                     starting segment for simulation process";
  if(seg.module_name_ != "")
    out << " in module " << seg.module_name_ <<"\n";
  if(seg.deref_pointer_read_)
    out << "                     contains pointer read\n";
  if(seg.deref_pointer_write_)
    out << "                     contains pointer write\n";
  if(seg.is_entrance_to_channel_method_)
    out << "                     Node for channel method\n";
  out << "\n                     */\n";
}

void
risc::sg::VertexPropertyWriter::replace_escape_characters(
    std::string &unparsed_stmt)
{
  boost::algorithm::replace_all(unparsed_stmt, "&",  "&amp;");
  boost::algorithm::replace_all(unparsed_stmt, "\"", "&quot;");
  boost::algorithm::replace_all(unparsed_stmt, "\'", "&#39;");
  boost::algorithm::replace_all(unparsed_stmt, "<",  "&lt;");
  boost::algorithm::replace_all(unparsed_stmt, ">",  "&gt;");
  boost::algorithm::replace_all(unparsed_stmt, "{",  "&#123;");
  boost::algorithm::replace_all(unparsed_stmt, "}",  "&#125;");
  boost::algorithm::replace_all(unparsed_stmt, "|",  "&#124;");
  boost::algorithm::replace_all(unparsed_stmt, "\\", "\\\\");
}

std::string
risc::sg::VertexPropertyWriter::write_pcp(
    PortCallPath &pcp)
{
  std::string result;

  if(pcp.path_.empty() && pcp.pcp_type_ == 0) {
    return result;
  }

  result = " [";
  if(pcp.pcp_type_ == 0) {
    bool first = true;
    for(std::list<SgFunctionCallExp*>::iterator
        port_call_iter  = pcp.path_.begin();
        port_call_iter != pcp.path_.end();
        port_call_iter++) 
    {
      std::string temp((*port_call_iter)->unparseToString());
      if(first) {
        first = false;
      } else {
        result += " -> ";
      }
      result += temp;
    }
  } else {
    SgVariableDefinition* pcp_port_of_func = pcp.port_;

    std::string temp = pcp_port_of_func->get_declaration()->
      get_decl_item(0)->get_qualified_name().getString();
    result += temp;
  }

  result += "]";
  replace_escape_characters(result);
  return result;
}

bool risc::sg::VertexPropertyWriter::print_read_write_access = false;

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
