// bug20180523.cc:
// derived from $SYSTEMC_HOME/examples/sysc/risc_cpu/pic.{h|cpp}
// adapted for RISC by DM, May 2018

#include "systemc.h"

struct pic : sc_module { 
  	sc_in<bool>  		ireq0; 		// interrupt request 0   	
  	sc_in<bool>  		ireq1;    	// interrupt request 1
  	sc_in<bool>  		ireq2;    	// interrupt request 2
  	sc_in<bool>  		ireq3;    	// interrupt request 3
  	sc_in<bool>  		cs;    		// chip select
  	sc_in<bool>  		rd_wr;    	// read or write
  	sc_in<bool>  		intack_cpu;    	// interrupt acknowledge from CPU
  	sc_out<bool>  		intreq;         // interrupt request to CPU		
  	sc_out<bool>  		intack;         // interrupt acknowledge to devices		
  	sc_out<unsigned>        vectno;         // vector number		

  //Constructor 
	SC_CTOR(pic) {
		SC_THREAD(entry);
	}

  // Process functionality in member function below
  void entry(){

	while(1) {
		// the following line results in a crash with RISC v0.4.1
		wait(ireq0.value_changed_event() | ireq1.value_changed_event() |
			ireq2.value_changed_event() | ireq3.value_changed_event());
		cout << "in entry() of pic" << " @" << sc_time_stamp() << "\n";
		if (ireq0.read() == true) {
			intreq.write(true);
			vectno.write(0);
		} else if (ireq1.read() == true) {
				intreq.write(true);
				vectno.write(1);
			} else if (ireq2.read() == true) {
					intreq.write(true);
					vectno.write(2);
				} else if (ireq3.read() == true) {
							intreq.write(true);
							vectno.write(2);
					} else {
					} 
		if ((intack_cpu.read() == true) && (cs.read() == true)) {
				intreq.write(false);
		}
	}
  }
};

struct stimulus : sc_module {
	sc_out<bool>  		ireq0; 		// interrupt request 0   	
  	sc_out<bool>  		ireq1;    	// interrupt request 1
  	sc_out<bool>  		ireq2;    	// interrupt request 2
  	sc_out<bool>  		ireq3;    	// interrupt request 3
  	sc_out<bool>  		cs;    		// chip select
  	sc_out<bool>  		rd_wr;    	// read or write
  	sc_out<bool>  		intack_cpu;    	// interrupt acknowledge from CPU

	SC_CTOR(stimulus) {
		SC_THREAD(stim_behav);
	}

	void stim_behav() {
	cout << "Initializing values in stim" << " @" << sc_time_stamp() << "\n";
	ireq0.write(false);
	ireq1.write(false);
	ireq2.write(false);
	ireq3.write(false);
	cs.write(false);
	rd_wr.write(false);
	intack_cpu.write(false);

	wait(5,SC_NS);
	cout << "Stimulus requesting interrupt 3" << " @" << sc_time_stamp() << "\n";
		ireq3.write(true);
		cout << "Stimulus finished requesting interrupt" << " @" << sc_time_stamp() << "\n";
	wait(5,SC_NS);
	}
};

struct monitor : sc_module {
  	sc_in<bool>  		intreq;         // interrupt request to CPU		
  	sc_in<bool>  		intack;         // interrupt acknowledge to devices		
  	sc_in<unsigned>         vectno;         // vector number		

	SC_CTOR(monitor) {
		SC_THREAD(mon_behav);
	}

	void mon_behav() {
		wait(vectno.value_changed_event());
		bool req = intreq.read();
		bool tack = intack.read();
		int vect = vectno.read();
		printf("sc_method mon_behav() detected change: intreq=%d intack=%d vectno=%d",req,tack,vect);
		cout << " @" << sc_time_stamp() << "\n";
	}
};

struct top : sc_module {
	stimulus stim;
	pic pi;
	monitor mon;

	sc_signal<bool>  		ireq0; 		// interrupt request 0   	
	sc_signal<bool>  		ireq1;    	// interrupt request 1
	sc_signal<bool>  		ireq2;    	// interrupt request 2
	sc_signal<bool>  		ireq3;    	// interrupt request 3
	sc_signal<bool>  		cs;    		// chip select
	sc_signal<bool>  		rd_wr;    	// read or write
	sc_signal<bool>  		intack_cpu;    	// interrupt acknowledge from CPU
	sc_signal<bool>  		intreq;         // interrupt request to CPU		
	sc_signal<bool>  		intack;         // interrupt acknowledge to devices		
	sc_signal<unsigned>		vectno;         // vector number		

	SC_CTOR(top):pi("pi"),stim("stim"),mon("mon") {
		stim.ireq0.bind(ireq0);
		stim.ireq1.bind(ireq1);
		stim.ireq2.bind(ireq2);
		stim.ireq3.bind(ireq3);
		stim.cs.bind(cs);
		stim.rd_wr.bind(rd_wr);
		stim.intack_cpu.bind(intack_cpu);
		mon.intreq.bind(intreq);
		mon.intack.bind(intack);
		mon.vectno.bind(vectno);
		pi.ireq0.bind(ireq0);
		pi.ireq1.bind(ireq1);
		pi.ireq2.bind(ireq2);
		pi.ireq3.bind(ireq3);
		pi.cs.bind(cs);
		pi.rd_wr.bind(rd_wr);
		pi.intack_cpu.bind(intack_cpu);
		pi.intreq.bind(intreq);
		pi.intack.bind(intack);
		pi.vectno.bind(vectno);
	}
};

int sc_main(int argc,char** argv) {
	top TOP("TOP");
	sc_start();
	return 0;
}
