// bug20180717b.cc
// (was bug_scoping.cpp, reported by DM 07/17/18)

#include <stdio.h>
#include "systemc.h"

SC_MODULE(TOP) {
	SC_CTOR(TOP) {
		SC_THREAD(behav);
	}

	void behav() {
		wait(sc_time(5,SC_NS));				// user code is here
  // (this) ->  wait (sc_time::sc_time(5,sc_core::SC_NS),1);	// code generated is this

// -> GCC 4.4.7 has no problem with it
// -> GCC 6.3.1 fails:
// risc_bug_scoping.cpp: In member function 'void TOP::behav()':
// risc_bug_scoping.cpp:24:55: error: cannot call constructor 'sc_core::sc_time::sc_time' directly [-fpermissive]
//      (this) ->  wait (sc_time::sc_time(5,sc_core::SC_NS),1);
//                                                             ^
// risc_bug_scoping.cpp:24:55: note: for a function-style cast, remove the redundant '::sc_time'
// -> GCC 6.3.1 can be told to accept this with -fpermissive
	}
};

int sc_main(int argc, char* argv[]) {
	TOP t("t");
	sc_start();
}

// UPDATE 07/24/18, RD:
// I've validated that the problem is the ROSE code generator.
// The problem is still present in the latest ROSE version 0.9.10.54.
// I've now reported the following test case to Dan Quinlan, hoping for a quick fix!

class Point
{ public:
  Point(int x, int y)
  : x(x), y(y)
  { }
  int x, y;
};

void PrintPoint(Point p)
{
   printf("(%d,%d)\n", p.x, p.y);
}

void Rectangle(Point p1, Point p2)
{
   printf("(%d,%d)x(%d,%d)\n",
	p1.x, p1.y, p2.x, p2.y);
}

int NOTmain(void)
{
  Point p(15,27);
  PrintPoint(p);			// works
  PrintPoint(Point(42,0));		// fails
  Rectangle(Point(1,2),Point(3,4));	// fails
  return 0;
}

// EOF
