// sensitive1.cc:
//
// 07/07/18 RD	regression test for threads with static sensitivity

#include <assert.h>
#include "systemc.h"

#define T 5

SC_MODULE(Top) {
	sc_event tick_even, tick_odd, tick_ten;
	int n_even, n_odd, n_ten;
	int work_even, work_odd, work_ten, work_twice_as_hard, work_tasks;
	sc_time tn1, tn2, tn10;
	sc_time tw1, tw2, tw10, tw2x, twt;

	SC_CTOR(Top) {
		n_even = n_odd = n_ten = 0;
		work_even = work_odd = work_ten = work_twice_as_hard = work_tasks = 0;
		tn1 = tn2 = tn10 = SC_ZERO_TIME;
		tw1 = tw2 = tw10 = tw2x = twt = SC_ZERO_TIME;
		SC_THREAD(clock_even);
		SC_THREAD(clock_odd);
		SC_THREAD(clock_ten);
		SC_THREAD(worker_even);
		sensitive << tick_even;
		SC_THREAD(worker_odd);
		sensitive << tick_odd;
		SC_THREAD(worker_ten);
		sensitive << tick_ten;
		SC_THREAD(worker_twice_as_hard);
		sensitive << tick_odd << tick_even;
		sensitive << tick_ten;
		//SC_THREAD(worker_tasks); // Two threads, same job. Conflicts!
		SC_THREAD(worker_tasks);
		sensitive << tick_odd << tick_even;
	}

	void clock_even() {
		for(int i=0; i<T; i+=1) {
			wait(2,SC_SEC);
			std::cout << sc_time_stamp().to_string() << ": clock even tick!\n";
			tick_even.notify(SC_ZERO_TIME);
			n_even++;
		}
		tn2 = sc_time_stamp();
	}

	void clock_odd() {
		wait(1,SC_SEC);
		for(int i=0; i<T; i+=1) {
			std::cout << sc_time_stamp().to_string() << ": clock odd tick!\n";
			tick_odd.notify(SC_ZERO_TIME);
			n_odd++;
			wait(2,SC_SEC);
		}
		tn1 = sc_time_stamp();
	}

	void clock_ten() {
		for(int i=0; i<T/5; i+=1) {
			wait(10,SC_SEC);
			std::cout << sc_time_stamp().to_string() << ": clock ten tick!\n";
			tick_ten.notify(SC_ZERO_TIME);
			n_ten++;
		}
		tn10 = sc_time_stamp();
	}

	void worker_even() {
		while(true) {
			wait();
			std::cout << sc_time_stamp().to_string() << ": worked even shift!\n";
			work_even++;
			tw2 = sc_time_stamp();
		}
	}

	void worker_odd() {
		while(true) {
			wait();
			std::cout << sc_time_stamp().to_string() << ": worked odd shift!\n";
			work_odd++;
			tw1 = sc_time_stamp();
		}
	}

	void worker_ten() {
		while(true) {
			wait();
			std::cout << sc_time_stamp().to_string() << ": worked ten shift!\n";
			work_ten++;
			tw10 = sc_time_stamp();
		}
	}

	void worker_twice_as_hard() {
		while(true) {
			wait();
			std::cout << sc_time_stamp().to_string() << ": worked another shift!\n";
			work_twice_as_hard++;
			tw2x = sc_time_stamp();
		}
		//task6(); // Two threads for task 6. Conflict!
	}

	void worker_tasks() {
		wait();
		std::cout << sc_time_stamp().to_string() << ": worked task 1.\n";
		work_tasks++;
		twt = sc_time_stamp();
		{	wait();
			std::cout << sc_time_stamp().to_string() << ": worked task 2.\n";
			work_tasks++;
			twt = sc_time_stamp();
		}
		for(int i=0; i<3; i++)
		{	wait();
			std::cout << sc_time_stamp().to_string() << ": worked on tasks 3,4,5.\n";
			work_tasks++;
			twt = sc_time_stamp();
		}
		task6();
		task7and8();
	}

	void task6() {
		wait();
		std::cout << sc_time_stamp().to_string() << ": worked task 6.\n";
		work_tasks++;
		twt = sc_time_stamp();
	}

	void task7and8() {
		wait();
		std::cout << sc_time_stamp().to_string() << ": worked task 7.\n";
		work_tasks++;
		twt = sc_time_stamp();
		task8();
	}

	void task8() {
		wait();
		std::cout << sc_time_stamp().to_string() << ": worked task 8.\n";
		work_tasks++;
		twt = sc_time_stamp();
	}
};

Top top("top");

int sc_main(int argc, char** argv) {
	const sc_time tn1(1+2*T, SC_SEC);
	const sc_time tn2(2*T, SC_SEC);
	const sc_time tn10(10*T/5, SC_SEC);
	const sc_time tw1(2*T-1, SC_SEC);
	const sc_time tw2(2*T, SC_SEC);
	const sc_time tw10(10*T/5, SC_SEC);
	const sc_time tw2x(2*T, SC_SEC);
	const sc_time twt(8, SC_SEC);

	cout << "sensitive1 test starts" << endl;
	sc_start();
	assert(top.n_odd == T);
	assert(top.n_even == T);
	assert(top.n_ten == T/5);
	assert(top.work_odd == T);
	assert(top.work_even == T);
	assert(top.work_ten == T/5);
	assert(top.work_twice_as_hard == 2*T);
	assert(top.work_tasks == 8);
	assert(top.tn1 == tn1);
	assert(top.tn2 == tn2);
	assert(top.tn10 == tn10);
	assert(top.tw1 == tw1);
	assert(top.tw2 == tw2);
	assert(top.tw10 == tw10);
	assert(top.tw2x == tw2x);
	assert(top.twt == twt);
	cout << "sensitive1 test done." << endl;
	return 0;
}

/* ex: set softtabstop=8 tabstop=8 shiftwidth=8: */
