#include "systemc.h"

///////////////////////////////// Interface  /////////////////////////////////

class Interface1: virtual public sc_interface {

public:
  #pragma RISC read1 unconditional-wait event
  virtual int read1() = 0;
  #pragma RISC write1 unconditional-wait event
  virtual void write1(const int &value) = 0;
};

class Interface2: virtual public sc_interface
{
public:
  virtual int read2() = 0;
  virtual void write2(const int &value) = 0;
};

class Interface3: virtual public sc_interface
{
public:
  #pragma RISC read3 unconditional-wait event
  virtual int read3() = 0;
  virtual void write3(const int &value) = 0;
};


///////////////////////////////// Interface  /////////////////////////////////
class Channel1:
  public sc_channel,
  public Interface1 {

public:
  Channel1(sc_module_name name):
    sc_channel(name)
  { }

  #pragma RISC read1 unconditional-wait event
  int read1()
  {
    return 1;
  }

  #pragma RISC write1 unconditional-wait event
  void write1(const int &value)
  { }
};

class Channel2:
  public sc_channel,
  public Interface2 {

public:
  Channel2(sc_module_name name):
    sc_channel(name)
  { }

  int read2()
  {
    return 2;
  }

  void write2(const int &value)
  { }
};

class Channel3:
  public sc_channel,
  public Interface3 {

public:
  Channel3(sc_module_name name):
    sc_channel(name)
  { }

  #pragma RISC read3 unconditional-wait event
  int read3()
  {
    return 3;
  }

  void write3(const int &value)
  { }
};

///////////////////////////////// Sender /////////////////////////////////

class Sender1: public sc_module {

public:

  SC_HAS_PROCESS(Sender1);

  Sender1(sc_module_name mn):
    sc_module(mn)
  {
    SC_THREAD(main);
  }

  sc_port<Interface1> port;

  void main()
  {
    port->write1(1);
  }
};

class Sender2: public sc_module {

public:

  SC_HAS_PROCESS(Sender2);

  Sender2(sc_module_name mn):
    sc_module(mn)
  {
    SC_THREAD(main);
  }

  sc_port<Interface2> port;

  void main()
  {
    port->write2(2);
  }
};

class Sender3: public sc_module {

public:

  SC_HAS_PROCESS(Sender3);

  Sender3(sc_module_name mn):
    sc_module(mn)
  {
    SC_THREAD(main);
  }

  sc_port<Interface3> port;

  void main()
  {
    port->write3(3);
  }
};

///////////////////////////////// Receiver /////////////////////////////////

class Receiver1: public sc_module {

public:

  SC_HAS_PROCESS(Receiver1);

  Receiver1(sc_module_name mn):
    sc_module(mn)
  {
    SC_THREAD(main);
  }

  sc_port<Interface1> port;

  void main()
  {
    port->read1();
  }
};

class Receiver2: public sc_module {

public:

  SC_HAS_PROCESS(Receiver2);

  Receiver2(sc_module_name mn):
    sc_module(mn)
  {
    SC_THREAD(main);
  }

  sc_port<Interface2> port;

  void main()
  {
    port->read2();
  }
};

class Receiver3: public sc_module {

public:

  SC_HAS_PROCESS(Receiver3);

  Receiver3(sc_module_name mn):
    sc_module(mn)
  {
    SC_THREAD(main);
  }

  sc_port<Interface3> port;

  void main()
  {
    port->read3();
  }
};

///////////////////////////////// Receiver /////////////////////////////////

class Top: public sc_module {

public:

  SC_HAS_PROCESS(Receiver3);

  Top(sc_module_name mn):
    sc_module(mn),
    c1("c1"),
    c2("c2"),
    c3("c3"),
    s1("s1"),
    s2("s2"),
    s3("s3"),
    r1("r1"),
    r2("r2"),
    r3("r3")
  { }

  Channel1 c1;
  Channel2 c2;
  Channel3 c3;

  Sender1 s1;
  Sender2 s2;
  Sender3 s3;

  Receiver1 r1;
  Receiver2 r2;
  Receiver3 r3;

  void before_end_of_elaboration()
  {
    s1.port.bind(c1);
    s2.port.bind(c2);
    s3.port.bind(c3);

    r1.port.bind(c1);
    r2.port.bind(c2);
    r3.port.bind(c3);
  }
};

int sc_main(int argc, char **argv)
{
  Top top("top");
  sc_start();
  return 0;
}

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
